## API for procedural generation of stereolitography files, a.k.a. STL.
## 
## Nim port of Py2STL from the Stereo Imagination suite.
## 
## Notes:
## 
## - generates plain text STL files that aren't exactly standard, but can be
##   parsed just fine by modern tools;
## - uses a right-handed coordinate system by default: you can interpret the y
##   axis as going into the screen and the z axis upwards, or else y as going
##   upwards and z as coming towards you (both can make sense);
## - all transformations are made around the origin: you might want to scale
##   and rotate your shapes before translating them into position.
## - all angles are given in radians -- make sure to convert from degrees if
##   necessary.
## 
## Nim2STL is deliberately very simple, not unlike a child's toy blocks;
## the whole point is not requiring a big, integrated system just to spit
## out a few shapes.

from math import sin, cos, PI
from strutils import join
import strformat

type Transform* = ref object of RootObj

method translateX*(self: Transform, x: float): Transform {.base.} = self
method translateY*(self: Transform, y: float): Transform {.base.} = self
method translateZ*(self: Transform, z: float): Transform {.base.} = self

method translate*(self: Transform; x, y, z: float): Transform {.base.} = self

method scaleX*(self: Transform, x: float): Transform {.base.} = self
method scaleY*(self: Transform, y: float): Transform {.base.} = self
method scaleZ*(self: Transform, z: float): Transform {.base.} = self

method scale*(self: Transform; x, y, z: float): Transform {.base.} = self

method rotateX*(self: Transform, x: float): Transform {.base.} = self
method rotateY*(self: Transform, y: float): Transform {.base.} = self
method rotateZ*(self: Transform, z: float): Transform {.base.} = self

method rotate*(self: Transform; x, y, z: float): Transform {.base.} = self

method shearX*(self: Transform; y, z: float): Transform {.base.} = self
method shearY*(self: Transform; x, z: float): Transform {.base.} = self
method shearZ*(self: Transform; x, y: float): Transform {.base.} = self

method distortX*(self: Transform; y, z: float): Transform {.base.} = self
method distortY*(self: Transform; x, z: float): Transform {.base.} = self
method distortZ*(self: Transform; x, y: float): Transform {.base.} = self

method `$`*(self: Transform): string {.base.} = "!"

type Vertex* = ref object of Transform
        x*, y*, z*: float

proc newVertex*(x = 0.0, y = 0.0, z = 0.0): Vertex = Vertex(x: x, y: y, z: z)

method translateX*(self: Vertex, x: float): Transform =
        self.x += x
        return self

method translateY*(self: Vertex, y: float): Transform =
        self.y += y
        return self

method translateZ*(self: Vertex, z: float): Transform =
        self.z += z
        return self

method translate*(self: Vertex; x, y, z: float): Transform =
        self.x += x
        self.y += y
        self.z += z
        return self

method scaleX*(self: Vertex, x: float): Transform =
        self.x *= x
        return self

method scaleY*(self: Vertex, y: float): Transform =
        self.y *= y
        return self

method scaleZ*(self: Vertex, z: float): Transform =
        self.z *= z
        return self

method scale*(self: Vertex; x, y, z: float): Transform =
        self.x *= x
        self.y *= y
        self.z *= z
        return self

method rotateX*(self: Vertex, x: float): Transform =
        let y = self.y
        let z = self.z
        self.y = y * cos(x) - z * sin(x)
        self.z = y * sin(x) + z * cos(x)
        return self

method rotateY*(self: Vertex, y: float): Transform =
        let x = self.x
        let z = self.z
        self.x = x * cos(y) - z * sin(y)
        self.z = x * sin(y) + z * cos(y)
        return self

method rotateZ*(self: Vertex, z: float): Transform =
        let x = self.x
        let y = self.y
        self.x = x * cos(z) - y * sin(z)
        self.y = x * sin(z) + y * cos(z)
        return self

method rotate*(self: Vertex; x, y, z: float): Transform =
        discard self.rotateX(x)
        discard self.rotateY(y)
        discard self.rotateZ(z)
        return self

method shearX*(self: Vertex; y, z: float): Transform =
        self.x += self.y * y + self.z * z
        return self

method shearY*(self: Vertex; x, z: float): Transform =
        self.y += self.x * x + self.z * z
        return self

method shearZ*(self: Vertex; x, y: float): Transform =
        self.z += self.x * x + self.y * y
        return self

method distortX*(self: Vertex; y, z: float): Transform =
        self.y *= abs(self.x) * y
        self.z *= abs(self.x) * z
        return self

method distortY*(self: Vertex; x, z: float): Transform =
        self.x *= abs(self.y) * x
        self.z *= abs(self.y) * z
        return self

method distortZ*(self: Vertex; x, y: float): Transform =
        self.x *= abs(self.z) * x
        self.y *= abs(self.z) * y
        return self

method `$`*(self: Vertex): string = "vertex {self.x} {self.y} {self.z}".fmt

type Facet* = seq[int]

type Shape* = ref object of Transform
        name*: string
        vertices*: seq[Vertex]
        faces*: seq[Facet]

proc newShape*(n = ""): Shape = Shape(name: n, vertices: @[], faces: @[])

method translateX*(self: Shape, x: float): Transform =
        for i in self.vertices:
                i.x += x
        return self

method translateY*(self: Shape, y: float): Transform =
        for i in self.vertices:
                i.y += y
        return self

method translateZ*(self: Shape, z: float): Transform =
        for i in self.vertices:
                i.z += z
        return self

method translate*(self: Shape; x, y, z: float): Transform =
        for i in self.vertices:
                discard i.translate(x, y, z)
        return self

method scaleX*(self: Shape, x: float): Transform =
        for i in self.vertices:
                i.x *= x
        return self

method scaleY*(self: Shape, y: float): Transform =
        for i in self.vertices:
                i.y *= y
        return self

method scaleZ*(self: Shape, z: float): Transform =
        for i in self.vertices:
                i.z *= z
        return self

method scale*(self: Shape; x, y, z: float): Transform =
        for i in self.vertices:
                discard i.scale(x, y, z)
        return self

method rotateX*(self: Shape, x: float): Transform =
        for i in self.vertices:
                discard i.rotateX(x)
        return self

method rotateY*(self: Shape, y: float): Transform =
        for i in self.vertices:
                discard i.rotateY(y)
        return self

method rotateZ*(self: Shape, z: float): Transform =
        for i in self.vertices:
                discard i.rotateZ(z)
        return self

method rotate*(self: Shape; x, y, z: float): Transform =
        for i in self.vertices:
                discard i.rotate(x, y, z)
        return self

method shearX*(self: Shape; y, z: float): Transform =
        for i in self.vertices:
                discard i.shearX(y, z)
        return self

method shearY*(self: Shape; x, z: float): Transform =
        for i in self.vertices:
                discard i.shearY(x, z)
        return self

method shearZ*(self: Shape; x, y: float): Transform =
        for i in self.vertices:
                discard i.shearZ(x, y)
        return self

method distortX*(self: Shape; y, z: float): Transform =
        for i in self.vertices:
                discard i.distortX(y, z)
        return self

method distortY*(self: Shape; x, z: float): Transform =
        for i in self.vertices:
                discard i.distortY(x, z)
        return self

method distortZ*(self: Shape; x, y: float): Transform =
        for i in self.vertices:
                discard i.distortZ(x, y)
        return self

method `$`*(self: Shape): string {.noSideEffect.} =
        result = "solid {self.name}\n".fmt
        for i in self.faces:
                result &= "facet normal 0 0 0\n"
                result &= "\touter loop\n"
                for j in i:
                        let v = self.vertices[j]
                        result &= "\t\t{v}\n".fmt
                result &= "\tendloop\n"
                result &= "endfacet\n"
        result &= "endsolid {self.name}".fmt

type Group* = ref object of Transform
        members*: seq[Transform]

proc newGroup*(m: varargs[Transform]): Group = Group(members: @m)

method translateX*(self: Group, x: float): Transform =
        for i in self.members:
                discard i.translateX(x)
        return self

method translateY*(self: Group, y: float): Transform =
        for i in self.members:
                discard i.translateY(y)
        return self

method translateZ*(self: Group, z: float): Transform =
        for i in self.members:
                discard i.translateZ(z)
        return self

method translate*(self: Group; x, y, z: float): Transform =
        for i in self.members:
                discard i.translate(x, y, z)
        return self

method scaleX*(self: Group, x: float): Transform =
        for i in self.members:
                discard i.scaleX(x)
        return self

method scaleY*(self: Group, y: float): Transform =
        for i in self.members:
                discard i.scaleY(y)
        return self

method scaleZ*(self: Group, z: float): Transform =
        for i in self.members:
                discard i.scaleZ(z)
        return self

method scale*(self: Group; x, y, z: float): Transform =
        for i in self.members:
                discard i.scale(x, y, z)
        return self

method rotateX*(self: Group, x: float): Transform =
        for i in self.members:
                discard i.rotateX(x)
        return self

method rotateY*(self: Group, y: float): Transform =
        for i in self.members:
                discard i.rotateY(y)
        return self

method rotateZ*(self: Group, z: float): Transform =
        for i in self.members:
                discard i.rotateZ(z)
        return self

method rotate*(self: Group; x, y, z: float): Transform =
        for i in self.members:
                discard i.rotate(x, y, z)
        return self

method shearX*(self: Group; y, z: float): Transform =
        for i in self.members:
                discard i.shearX(y, z)
        return self

method shearY*(self: Group; x, z: float): Transform =
        for i in self.members:
                discard i.shearY(x, z)
        return self

method shearZ*(self: Group; x, y: float): Transform =
        for i in self.members:
                discard i.shearZ(x, y)
        return self

method distortX*(self: Group; y, z: float): Transform =
        for i in self.members:
                discard i.distortX(y, z)
        return self

method distortY*(self: Group; x, z: float): Transform =
        for i in self.members:
                discard i.distortY(x, z)
        return self

method distortZ*(self: Group; x, y: float): Transform =
        for i in self.members:
                discard i.distortZ(x, y)
        return self

method `$`*(self: Group): string = self.members.join("\n") & "\n"

proc add*(self: Group, member: Transform) = self.members.add(member)
proc `+=`*(self: Group, member: Transform) = self.members.add(member)

func sheet*(width=2.0, height=2.0): Shape =
        ## Make a simple rectangle in the x-y plane, centered on origin.
        result = Shape()
        result.vertices.add(newVertex(-1, -1, 0))
        result.vertices.add(newVertex(1, -1, 0))
        result.vertices.add(newVertex(1, 1, 0))
        result.vertices.add(newVertex(-1, 1, 0))
        result.faces.add(@[0, 1, 2])
        result.faces.add(@[0, 2, 3])
        discard result.scale(width / 2, height / 2, 1)

func cone*(h=2.0, r=1.0, segments=12): Shape =
        ## An uncapped cone of given height and radius, pointed along Z.
        result = Shape()
        result.vertices.add(newVertex(0, 0, h / 2.0))
        let step = 2.0 * PI / float(segments)
        var i = 0.0
        while i < 2.0 * PI:
                let v = newVertex(cos(i) * r, sin(i) * r, -h / 2.0)
                result.vertices.add(v)
                i += step
        for j in 1 ..< len(result.vertices) - 1:
                result.faces.add(@[0, j, j + 1])
        result.faces.add(@[0, len(result.vertices) - 1, 1])

func disc*(radius=1.0, segments=12): Shape =
        ## Make a disc of given radius -- basically a many-sided polygon.
        ## 
        ## Implemented as a perfectly flat cone.
        return cone(0, radius, segments)

func bucket*(h=2.0, r1=1.0, r2=1.0, segments=12): Shape =
        ## Make a truncated cone of given height and radii, open at both ends.
        result = Shape()
        let step = 2.0 * PI / float(segments)
        var i = 0.0
        while i < 2.0 * PI:
                let v = newVertex(cos(i) * r2, sin(i) * r2, -h / 2.0)
                result.vertices.add(v)
                i += step
        i = 0.0
        while i < 2.0 * PI:
                let v = newVertex(cos(i) * r1, sin(i) * r1, h / 2)
                result.vertices.add(v)
                i += step
        for j in 0 ..< segments - 1:
                result.faces.add(@[j, j + 1, j + 1 + segments])
                result.faces.add(@[j, j + 1 + segments, j + segments])
        result.faces.add(@[segments - 1, 0, segments])
        result.faces.add(@[segments - 1, segments, segments * 2 - 1])

func cylinder*(height=2.0, radius=1.0, segments=12): Shape =
        ## Make a bucket (truncated cone) with the same radius at both ends.
        return bucket(height, radius, radius, segments)

func prism*(height=2.0, radius1=1.0, radius2=1.0): Shape =
        ## Make a bucket with just three segments, capped at both ends.
        result = bucket(height, radius1, radius2, 3)
        result.faces.add(@[0, 2, 1])
        result.faces.add(@[3, 4, 5])

func simplex*(height=2.0, radius=1.0): Shape =
        ## Make a tetrahedron (a cone with just three segments and a bottom).
        result = cone(height, radius, 3)
        result.faces.add(@[1, 3, 2])

func pyramid*(x=2.0, y=2.0, z=2.0): Shape =
        ## Make a pyramid of given size (with a bottom).
        result = Shape()

        result.vertices.add(newVertex(0, 0, 1))

        result.vertices.add(newVertex(-1, -1, -1))
        result.vertices.add(newVertex(1, -1, -1))
        result.vertices.add(newVertex(1, 1, -1))
        result.vertices.add(newVertex(-1, 1, -1))

        result.faces.add(@[0, 1, 2])
        result.faces.add(@[0, 2, 3])
        result.faces.add(@[0, 3, 4])
        result.faces.add(@[0, 4, 1])
        # Bottom side
        result.faces.add(@[1, 4, 3])
        result.faces.add(@[1, 3, 2])

        discard result.scale(x / 2.0, y / 2.0, z / 2.0)

func octahedron*(x=2.0, y=2.0, z=2.0): Shape =
        ## Make an octahedron of given width, depth and height.
        result = Shape()

        result.vertices.add(newVertex(0, 0, 1))

        result.vertices.add(newVertex(-1, -1, 0))
        result.vertices.add(newVertex(1, -1, 0))
        result.vertices.add(newVertex(1, 1, 0))
        result.vertices.add(newVertex(-1, 1, 0))

        result.vertices.add(newVertex(0, 0, -1))

        # Top half
        result.faces.add(@[0, 1, 2])
        result.faces.add(@[0, 2, 3])
        result.faces.add(@[0, 3, 4])
        result.faces.add(@[0, 4, 1])
        # Bottom half
        result.faces.add(@[5, 2, 1])
        result.faces.add(@[5, 3, 2])
        result.faces.add(@[5, 4, 3])
        result.faces.add(@[5, 1, 4])

        discard result.scale(x / 2.0, y / 2.0, z / 2.0)

func box*(x=2.0, y=2.0, z=2.0): Shape =
        ## Make a cuboid of given width, depth and height.
        result = Shape()

        result.vertices.add(newVertex(-1, -1, -1))
        result.vertices.add(newVertex(-1, 1, -1))
        result.vertices.add(newVertex(1, 1, -1))
        result.vertices.add(newVertex(1, -1, -1))

        result.vertices.add(newVertex(-1, -1, 1))
        result.vertices.add(newVertex(1, -1, 1))
        result.vertices.add(newVertex(1, 1, 1))
        result.vertices.add(newVertex(-1, 1, 1))

        # Bottom side
        result.faces.add(@[0, 1, 3])
        result.faces.add(@[1, 2, 3])
        # Top side
        result.faces.add(@[4, 5, 7])
        result.faces.add(@[5, 6, 7])
        # Front side
        result.faces.add(@[0, 3, 4])
        result.faces.add(@[3, 5, 4])
        # Back side
        result.faces.add(@[1, 7, 2])
        result.faces.add(@[2, 7, 6])
        # Left side
        result.faces.add(@[0, 4, 1])
        result.faces.add(@[1, 4, 7])
        # Right side
        result.faces.add(@[2, 6, 5])
        result.faces.add(@[2, 5, 3])

        discard result.scale(x / 2.0, y / 2.0, z / 2.0)

when isMainModule:
        let t: Transform = newVertex(1, 2, 3)
        echo "Vertex before rotation: ", t
        discard t.rotateZ(PI / 2)
        echo "Vertex after rotation: ", t

        echo sheet()
