## Load STL or OBJ files and render them as display-independent wireframe.
## 
## Rendering is abstracted: all it does is pass line coordinates to a callback.

import strutils
import tables
import nim2stl

type Camera* = object
        x*, y*, z*, f*: float
        px*, py*, scale*: float

proc project*(self: var Camera; x, y, z: float) =
        let x = x - self.x
        let y = y - self.y
        let z = z - self.z
        if z != 0:
                self.scale = self.f / z
                self.px = x * self.scale
                self.py = y * self.scale
        else:
                self.scale = 0
                self.px = 0
                self.py = 0

func `+`*(v1, v2: Vertex): Vertex =
        newVertex(v1.x + v2.x, v1.y + v2.y, v1.z + v2.z)

func `-`*(v1, v2: Vertex): Vertex =
        newVertex(v1.x - v2.x, v1.y - v2.y, v1.z - v2.z)

func `/`*(v: Vertex, n: float): Vertex =
        newVertex(v.x / n, v.y / n, v.z / n)

type Renderer* = proc(x1, y1, x2, y2: float)

method renderBy*(self: Transform, cam: var Camera, rend: Renderer) {.base.} =
        discard 

method renderBy*(self: Shape, cam: var Camera, rend: Renderer) =
        var proj: seq[(float, float, bool)]
        for i in self.vertices:
                # Convert right-hand coords to left hand for the camera.
                cam.project(i.x, i.z, i.y)
                proj.add((cam.px, cam.py, i.y > cam.z + 0.001))
        for i in self.faces:
                if len(i) == 3:
                        let v0 = proj[i[0]]
                        let v1 = proj[i[1]]
                        let v2 = proj[i[2]]
                        # Skip edges with an end behind the camera.
                        if v0[2] and v1[2]:
                                rend(v0[0], v0[1], v1[0], v1[1])
                        if v1[2] and v2[2]:
                                rend(v1[0], v1[1], v2[0], v2[1])
                        if v2[2] and v0[2]:
                                rend(v2[0], v2[1], v0[0], v0[1])
                else: # Assume a quad to avoid testing again.
                        let v0 = proj[i[0]]
                        let v1 = proj[i[1]]
                        let v2 = proj[i[2]]
                        let v3 = proj[i[3]]
                        # Skip edges with an end behind the camera.
                        if v0[2] and v1[2]:
                                rend(v0[0], v0[1], v1[0], v1[1])
                        if v1[2] and v2[2]:
                                rend(v1[0], v1[1], v2[0], v2[1])
                        if v2[2] and v3[2]:
                                rend(v2[0], v2[1], v3[0], v3[1])
                        if v3[2] and v0[2]:
                                rend(v3[0], v3[1], v0[0], v0[1])

method renderBy*(self: Group, cam: var Camera, rend: Renderer) =
        for i in self.members:
                i.renderBy(cam, rend)

method minBound*(self: Transform): Vertex {.base.} = nil

method minBound*(self: Vertex): Vertex = newVertex(self.x, self.y, self.z)

method minBound*(self: Shape): Vertex {.noSideEffect.} =
        if len(self.vertices) == 0:
                return newVertex()
        else:
                var min_x = Inf
                var min_y = Inf
                var min_z = Inf
                for i in self.vertices:
                        if i.x < min_x:
                                min_x = i.x
                        if i.y < min_y:
                                min_y = i.y
                        if i.z < min_z:
                                min_z = i.z
                return newVertex(min_x, min_y, min_z)

method minBound*(self: Group): Vertex {.noSideEffect.} =
        if len(self.members) == 0:
                return newVertex(0, 0, 0)
        else:
                var min_x = Inf
                var min_y = Inf
                var min_z = Inf
                for i in self.members:
                        let bb_min = i.minBound
                        if bb_min.x < min_x:
                                min_x = bb_min.x
                        if bb_min.y < min_y:
                                min_y = bb_min.y
                        if bb_min.z < min_z:
                                min_z = bb_min.z
                return newVertex(min_x, min_y, min_z)

method maxBound*(self: Transform): Vertex {.base.} = nil

method maxBound*(self: Vertex): Vertex = newVertex(self.x, self.y, self.z)

method maxBound*(self: Shape): Vertex {.noSideEffect.} =
        if len(self.vertices) == 0:
                return newVertex()
        else:
                var max_x = NegInf
                var max_y = NegInf
                var max_z = NegInf
                for i in self.vertices:
                        if i.x > max_x:
                                max_x = i.x
                        if i.y > max_y:
                                max_y = i.y
                        if i.z > max_z:
                                max_z = i.z
                return newVertex(max_x, max_y, max_z)

method maxBound*(self: Group): Vertex {.noSideEffect.} =
        if len(self.members) == 0:
                return newVertex(0, 0, 0)
        else:
                var max_x = NegInf
                var max_y = NegInf
                var max_z = NegInf
                for i in self.members:
                        let bb_max = i.maxBound
                        if bb_max.x > max_x:
                                max_x = bb_max.x
                        if bb_max.y > max_y:
                                max_y = bb_max.y
                        if bb_max.z > max_z:
                                max_z = bb_max.z
                return newVertex(max_x, max_y, max_z)

method vertexCount*(self: Transform): int {.base, noSideEffect.} = 0

method vertexCount*(self: Vertex): int {.noSideEffect.} = 1

method vertexCount*(self: Shape): int {.noSideEffect.} = len(self.vertices)

method vertexCount*(self: Group): int {.noSideEffect.} =
        for i in self.members:
                result += i.vertexCount

method faceCount*(self: Transform): int {.base, noSideEffect.} = 0

method faceCount*(self: Vertex): int {.noSideEffect.} = 0

method faceCount*(self: Shape): int {.noSideEffect.} = len(self.faces)

method faceCount*(self: Group): int {.noSideEffect.} =
        for i in self.members:
                result += i.faceCount

proc objFromFile*(source: string): Shape =
        result = Shape()
        for i in source.lines:
                var fields = i.splitWhitespace
                if len(fields) == 0:
                        continue
                elif fields[0] == "v":
                        let x = parseFloat(fields[1])
                        let y = parseFloat(fields[2])
                        let z = parseFloat(fields[3])
                        result.vertices.add(newVertex(x, y, z))
                elif fields[0] == "f":
                        var face: seq[int]
                        for j in 1 ..< len(fields):
                                let vertex = fields[j].split('/')
                                face.add(parseInt(vertex[0]) - 1)
                        result.faces.add(face)

type StlParser* = object
        source*: seq[string]
        buffer*: string
        index*: int
        vertices: Table[string, int]

proc newStlParser*(lines: seq[string]): StlParser =
        result = StlParser(source: lines)
        if len(result.source) > 0:
                result.buffer = result.source[0].strip
                result.index = 1

proc hasMore*(self: StlParser): bool = self.index < len(self.source)

proc advanceLine*(self: var StlParser) =
        if self.hasMore:
                self.buffer = self.source[self.index].strip
                inc self.index

proc matchStart*(self: var StlParser, text: string): bool =
        if not self.hasMore:
                return false
        elif self.buffer.startsWith(text):
                self.advanceLine
                return true
        else:
                return false

proc matchLine*(self: var StlParser, text: string): bool =
        if self.buffer == text:
                self.advanceLine
                return true
        else:
                return false

proc parseVertex*(self: var StlParser, shape: var Shape): int =
        if self.buffer in self.vertices:
                return self.vertices[self.buffer]
        else:
                let fields = self.buffer.splitWhitespace
                let x = parseFloat(fields[1])
                let y = parseFloat(fields[2])
                let z = parseFloat(fields[3])
                let v = newVertex(x, y, z)
                shape.vertices.add(v)
                let idx = len(shape.vertices) - 1
                self.vertices[self.buffer] = idx
                return idx

proc parseFacet*(self: var StlParser, shape: var Shape) =
        if not self.hasMore:
                raise newException(ValueError,
                        "Unexpected end of input in solid")
        elif not self.matchStart("facet"):
                raise newException(ValueError, "STL facet expected")
        elif not self.matchLine("outer loop"):
                raise newException(ValueError, "STL outer loop expected")

        while not self.matchLine("endloop"):
                var facet: Facet
                while self.buffer.startsWith("vertex"):
                        facet.add(self.parseVertex(shape))
                        self.advanceLine
                shape.faces.add(facet)
                if self.buffer != "endloop":
                        raise newException(ValueError,
                                "STL endloop expected")

        if not self.matchLine("endfacet"):
                raise newException(ValueError, "STL endfacet expected")

proc parseSolid*(self: var StlParser, shape: var Shape): bool =
        if not self.hasMore:
                return false
        elif self.matchStart("solid"):
                self.vertices.clear
                while not self.matchStart("endsolid"):
                        self.parseFacet(shape)
                return true
        else:
                raise newException(ValueError, "STL solid expected")

when isMainModule:
        echo Camera(f: 200)
        let g = newGroup(box())
        echo "bounding box goes from ", g.minBound, " to ", g.maxBound
        echo g.vertexCount, " vertices and ", g.faceCount, " faces found."
