from math import degToRad
from os import extractFilename
import strformat, strutils

import nigui

import nim2stl, wire3d

const DarkSlateGray = rgb(47, 79, 79)
const SeaShell = rgb(255, 245, 238)
const GhostWhite = rgb(248, 248, 255)

const HelpText = """
WireView 3D release 2021-04-23, by No Time To Play
Simple viewer for 3D models in STL or OBJ format.

Keys:

    F1    show this message
    O     open file from disk
    S     save file as STL
    E     export a screenshot
    +     zoom in
    -     zoom out
    R     reset view
    Q     quit program
    
    arrows to rotate

WireView 3D is open source under the MIT License.
"""

app.init

var top = newWindow("WireView 3D")

top.width = 854.scaleToDpi
top.height = 480.scaleToDpi

var layout = newLayoutContainer(Layout_Vertical)

layout.widthMode = WidthMode_Fill
layout.heightMode = HeightMode_Fill

var toolbar = newLayoutContainer(Layout_Horizontal)
var toOpen = newButton("Open...")
var toSave = newButton("Save...")
var toExport = newButton("Export...")
var toZoomIn = newButton("Zoom In")
var toZoomOut = newButton("Zoom Out")
var toLeft = newButton("Left")
var toDown = newButton("Down")
var toUp = newButton("Up")
var toRight = newButton("Right")
var toReset = newButton("Reset")

toolbar.add toOpen
toolbar.add toSave
toolbar.add toExport
toolbar.add toZoomIn
toolbar.add toZoomOut
toolbar.add toLeft
toolbar.add toDown
toolbar.add toUp
toolbar.add toRight
toolbar.add toReset

var viewport = newControl()

viewport.widthMode = WidthMode_Fill
viewport.heightMode = HeightMode_Fill

var status = newLabel("Ready")

layout.add toolbar
layout.add viewport
layout.add status

top.add layout

var askOpen = newOpenFileDialog()
askOpen.title = "Open 3D model"

var askSave = newSaveFileDialog()
askSave.title = "Save As STL"

var askExport = newSaveFileDialog()
askExport.title = "Export as image"

func sampleScene(): Group

var camera = Camera(f: float(viewport.height))
var model = sampleScene()

var boxMin = model.minBound
var boxMax = model.maxBound
var boxSize = boxMax - boxMin
var boxCenter = (boxMin + boxMax) / 2.0

camera.x = boxCenter.x
camera.y = boxCenter.z

camera.z = boxCenter.y - max(boxSize.x, boxSize.z) * 1.5
if camera.z >= boxMin.y:
        camera.z = boxCenter.y - boxSize.y

var zoomStep = abs(camera.z - boxCenter.y) / 5

var angleX = 0.0
var angleZ = 0.0

const rotStep = 15.0

func sampleScene(): Group =
        func makeTree(): Group =
                result = newGroup()
                result += cylinder().translateZ(1).scale(0.2, 0.2, 1)
                result += cone().translateZ(1).scaleZ(1.5).translateZ(2)

        let house = newGroup(box().translateZ(1))
        let roof = cone(2, 1, 4).translateZ(3)
        house += roof.scale(2, 2, 1).rotateZ(degToRad(45.0))

        result = newGroup()
        result += disc().scale(5, 5, 1)
        result += makeTree().translate(3, 1, 0)
        result += makeTree().translate(-1, -2, 0)
        result += makeTree().translate(-3, 3, 0)
        result += house.scale(0.5, 0.5, 0.5)

proc setupView() =
        boxMin = model.minBound
        boxMax = model.maxBound
        boxSize = boxMax - boxMin
        boxCenter = (boxMin + boxMax) / 2.0

        camera.x = boxCenter.x
        camera.y = boxCenter.z

        camera.z = boxCenter.y - max(boxSize.x, boxSize.z) * 1.5
        if camera.z >= boxMin.y:
                camera.z = boxCenter.y - boxSize.y

        zoomStep = abs(camera.z - boxCenter.y) / 5

        angleX = 0.0
        angleZ = 0.0

func position(c: Camera): string = "x:{c.x} y:{c.z} z: {c.y}".fmt

proc renderModel(canvas: Canvas) =
        camera.f = min(canvas.width, canvas.height).float

        canvas.textColor = SeaShell
        canvas.fontSize = canvas.height / 20
        canvas.areaColor = DarkSlateGray
        canvas.fill
        canvas.lineColor = GhostWhite
        
        let m2 = model.deepCopy
        let centerX = canvas.width / 2
        let centerY = canvas.height / 2
        
        discard m2.rotateZ(degToRad(angleZ)).rotateX(degToRad(angleX))
        m2.renderBy camera, proc (x1, y1, x2, y2: float) =
                let x1 = int(x1 + centerX)
                let y1 = int(centerY - y1)
                let x2 = int(x2 + centerX)
                let y2 = int(centerY - y2)
                canvas.drawLine x1, y1, x2, y2
        
        let fs = int(canvas.fontSize)
        
        canvas.drawText camera.position, fs, fs
        canvas.drawText "u:{angleZ} v:{angleX}".fmt, fs, fs * 2

func stats(m: Transform): string =
        "{m.vertexCount} vertices and {m.faceCount} faces loaded.".fmt

proc doOpen() =
        askOpen.run()
        if len(askOpen.files) == 0:
                status.text = "Open canceled."
        elif askOpen.files[0].toLowerAscii.endsWith(".stl"):
                try:
                        let source = readFile(askOpen.files[0])
                        var parser = newStlParser(source.splitLines)
                        var s = newShape()
                        model.members.setLen 0
                        while parser.parseSolid(s):
                                model += s
                                s = newShape()
                        top.title = extractFilename(askOpen.files[0])
                        status.text = model.stats
                        setupView()
                except IOError:
                        top.alert(getCurrentExceptionMsg(), "WireView 3D")
                except ValueError:
                        top.alert(getCurrentExceptionMsg(), "WireView 3D")
        elif askOpen.files[0].toLowerAscii.endsWith(".obj"):
                model.members.setLen 0
                model += objFromFile(askOpen.files[0])
                top.title = extractFilename(askOpen.files[0])
                status.text = model.stats
                setupView()
        else:
                top.alert("Only .stl and .obj files, please.", "WireView 3D")

proc doSave() =
        askSave.run()
        if askSave.file != "":
                try:
                        # The STL parser expects a final newline.
                        writeFile(askSave.file, $model & "\n")
                        top.title = extractFilename(askSave.file)
                        status.text = "File saved."
                except IOError:
                        top.alert(getCurrentExceptionMsg(), "WireView 3D")
        else:
                status.text = "Save canceled."

proc doExport() =
        askExport.run()
        if askExport.file != "":
                let screenshot = newImage()
                screenshot.resize(viewport.width, viewport.height)
                screenshot.canvas.renderModel
                try:
                        let fn = askExport.file.toLowerAscii
                        if fn.endsWith(".bmp"):
                                screenshot.saveToBitmapFile(askExport.file)
                                status.text = "Exported as " & askExport.file
                        elif fn.endsWith(".png"):
                                screenshot.saveToPngFile(askExport.file)
                                status.text = "Exported as " & askExport.file
                        elif fn.endsWith(".jpg") or fn.endsWith(".jpeg"):
                                screenshot.saveToJpegFile(askExport.file)
                                status.text = "Exported as " & askExport.file
                        else:
                                let msg = "Please save as BMP, PNG or JPG."
                                top.alert(msg, "WireView 3D")
                except IOError:
                        top.alert(getCurrentExceptionMsg(), "WireView 3D")
        else:
                status.text = "Export canceled."

toOpen.onClick = proc(event: ClickEvent) = doOpen()
toSave.onClick = proc(event: ClickEvent) = doSave()
toExport.onClick = proc(event: ClickEvent) = doExport()

toZoomIn.onClick = proc(event: ClickEvent) =
        if camera.z < boxCenter.y:
                camera.z += zoomStep
        viewport.forceRedraw

toZoomOut.onClick = proc(event: ClickEvent) =
        camera.z -= zoomStep
        viewport.forceRedraw

toLeft.onClick = proc(event: ClickEvent) =
        angleZ += rotStep
        if angleZ > 180.0:
                angleZ -= 360.0
        viewport.forceRedraw

toRight.onClick = proc(event: ClickEvent) =
        angleZ -= rotStep
        if angleZ < -180.0:
                angleZ += 360.0
        viewport.forceRedraw

toUp.onClick = proc(event: ClickEvent) =
        if angleX < 90.0:
                angleX += rotStep
        viewport.forceRedraw

toDown.onClick = proc(event: ClickEvent) =
        if angleX > -90.0:
                angleX -= rotStep
        viewport.forceRedraw

toReset.onClick = proc(event: ClickEvent) =
        angleX = 0.0
        angleZ = 0.0

        camera.z = boxCenter.y - max(boxSize.x, boxSize.z) * 1.5
        if camera.z >= boxMin.y:
                camera.z = boxCenter.y - boxSize.y
                
        viewport.forceRedraw

viewport.onDraw = proc (event: DrawEvent) = event.control.canvas.renderModel

top.onKeyDown = proc(event: KeyboardEvent) =
        case event.key
        of Key_F1:
                top.alert(HelpText, "WireView 3D")
        of Key_O:
                doOpen()
        of Key_Equal, Key_Plus, Key_NumpadAdd:
                if camera.z < boxCenter.y:
                        camera.z += zoomStep
                viewport.forceRedraw
        of Key_Minus, Key_NumpadSubtract:
                camera.z -= zoomStep
                viewport.forceRedraw
        of Key_Left, Key_Numpad4:
                angleZ += rotStep
                if angleZ > 180.0:
                        angleZ -= 360.0
                viewport.forceRedraw
        of Key_Right, Key_Numpad6:
                angleZ -= rotStep
                if angleZ < -180.0:
                        angleZ += 360.0
                viewport.forceRedraw
        of Key_Up, Key_Numpad8:
                if angleX < 90.0:
                        angleX += rotStep
                viewport.forceRedraw
        of Key_Down, Key_Numpad2:
                if angleX > -90.0:
                        angleX -= rotStep
                viewport.forceRedraw
        of Key_R:
                angleX = 0.0
                angleZ = 0.0

                camera.z = boxCenter.y - max(boxSize.x, boxSize.z) * 1.5
                if camera.z >= boxMin.y:
                        camera.z = boxCenter.y - boxSize.y
                        
                viewport.forceRedraw
        of Key_Q:
                app.quit
        else:
                discard

top.show
app.run
